import datetime
import os
from itertools import chain, starmap


def dict_compare(old_dict, new_dict, nested=None):
    """ Compare two dictionaries
        Only 1 level, ignoring attributes starting with '_'
    """
    key_prefix = nested + '|' if nested else ''
    intersect_keys = old_dict.keys() & new_dict.keys()
    modified = {key_prefix + k: dict(old=old_dict[k], new=new_dict[k], action='mod') for k in intersect_keys
                        if k[0] != '_'
                        and old_dict[k] != new_dict[k]
                        and not (isinstance(old_dict[k], dict) and isinstance(new_dict[k], dict))}

    nested_keys = [k for k in intersect_keys if k[0] != '_' and isinstance(old_dict[k], dict) and isinstance(new_dict[k], dict)]
    for k in nested_keys:
        x = dict_compare(old_dict[k], new_dict[k], key_prefix + k)
        if x:
            modified.update(x)

    added = new_dict.keys() - old_dict.keys()
    modified.update({key_prefix + k: dict(new=new_dict[k], action='add') for k in added if k[0] != '_'})

    deleted = old_dict.keys() - new_dict.keys()
    modified.update({key_prefix + k: dict(old=old_dict[k], action='del') for k in deleted if k[0] != '_'})
    if modified:
        return modified


def running_in_gcf():
    """ Determine if code is running in GCF using GCP_PROJECT
    """
    return os.getenv('GCP_PROJECT') is not None


def default_object(o):
    """ Default handler for json.dumps()"""
    if isinstance(o, (datetime.date, datetime.datetime)):
        return o.isoformat()
    raise TypeError("Type %s not serializable" % type(o))


# def buid_dict(composed_key, key_value):
#     def recursive_buid_dict(keys, value, nest_level=0):
#         if nest_level < len(keys):
#             return recursive_buid_dict(keys, value={keys[nest_level]: value}, nest_level=nest_level + 1)
#         elif nest_level == len(keys):
#             return value
#
#     return recursive_buid_dict(keys=list(reversed(composed_key.split('.'))), value=key_value)
#
# from collections import ChainMap
#
# def merge_dicts(dict_a, dict_b):
#     new_dict = dict()
#     for key in ChainMap(dict_a, dict_b):
#         val_a , val_b = dict_a.get(key), dict_b.get(key)
#         if isinstance(val_a, dict) and isinstance(val_b, dict):
#             new_dict[key] = merge_dicts(val_a, val_b)
#         elif val_a is None and val_b is None:
#             continue
#         elif val_a is None:
#             new_dict[key] = val_b
#         elif val_b is None or val_a == val_b:
#             new_dict[key] = val_a
#         else:
#             new_dict[key] = [val_a, val_b]
#     return new_dict

class Compare(object):

    def __init__(self, old_dictionary=None, new_dictionary=None, ignore_starting=None, ignore_ending=None):
        self.old_dict = old_dictionary
        self.new_dict = new_dictionary
        self.ignore_starting = ignore_starting
        self.ignore_ending = ignore_ending
        self.result = self.deepcompare(old_dict=old_dictionary,
                                       new_dict=new_dictionary,
                                       ignore_starting=self.ignore_starting,
                                       ignore_ending=self.ignore_starting)
        self.flatresult = self.flat()

    @staticmethod
    def intersect_lists(left_list, right_list):
        """Intersect two list and return difference as tuple (left, middle, right)

        USAGE:  
            l, _, _ = intersect_list(left_list, right_list)
                return list of values in left list not matched with right list
            
            _, m, _ = intersect_list(left_list, right_list)
                return list of values in matched in left list and right list
            
            _, _, r = intersect_list(left_list, right_list)
                return list of values in right list not matched with left list
            
            l, m, r = intersect_list(left_list, right_list)
                return all three above mentioned results

        EXAMPLE:
            l, m, r = intersect_list([1, 2, 3], [2, 3, 4])
            l = [1]
            m = [2, 3]
            r = [4]
        """
        l = [value for value in left_list if value not in right_list]
        m = [value for value in left_list if value in right_list]
        r = [value for value in right_list if value not in left_list]
        return l, m, r

    @staticmethod
    def unnest(dictionary, separator='.'):
        """Flatten a nested dictionary structure
        
        EXAMPLE:
        test_unnest = {'a': {'b': {'c': {'d': ['e', 'f'], 'g': 5}}}}
        unnest(test_unnest)
            returns {'a.b.c.d': ['e', 'f'], 'a.b.c.g': 5}
        """

        def unpack(parent_key, parent_value):
            """Unpack one level of nesting in a dictionary"""
            try:
                for key, value in parent_value.items():
                    yield (f'{parent_key}{separator}{key}', value)
            except AttributeError:
                # parent_value was not a dict, no need to flatten
                yield (f'{parent_key}', parent_value)

        if not isinstance(dictionary, dict):
            return dictionary
        while any(isinstance(value, dict) for value in dictionary.values()): # TODO secure breaking
            # Keep unpacking the dictionary until all value's are not dictionary's
            try:
                dictionary = dict(chain.from_iterable(starmap(unpack, dictionary.items())))
            except AttributeError:
                break
        return dictionary

    @staticmethod
    def exclude_keys(set_of_keys, starting=None, ending=None):
        """Remove keys from set of keys starting and/or ending with string
        :param set_of_keys: a set of string keys to be iterated for ignoring keys
        :param starting: keys starting with string to be removed from a set of keys (default None)
        :param ending: keys ending with string to be removed from a set of keys (default None)
        """

        try:
            clean_keys = set()
            for key in set_of_keys:
                if starting and ending and key.startswith(starting) and key.endswith(ending):
                    continue
                elif starting and key.startswith(starting):
                    continue
                elif ending and key.endswith(ending):
                    continue
                else:
                    clean_keys.add(key)
            return clean_keys
        except TypeError as te:
            raise te

    def listtodict(self, lst):
        """Convert list of dictionaries to dictionary of dictionaries using index of list
        converts list of objects to dictionary of objects
        :param lst: list or dictionary to be converted to dictionary of objects
        """
        #NOTE: if object is not a dictionary it will be converted using default key as parent and index as child key
        if not isinstance(lst, (list, dict)):
            return lst
        try:
            result = {key: self.listtodict(value) for key, value in lst.items()}
        except AttributeError:
            result = {f'{index}': self.listtodict(value) for index, value in enumerate(lst)}
        return result

    def deepcompare(self, old_dict, new_dict, ignore_starting=None, ignore_ending=None):
        """Compare nested dictionaries, including lists recursively
        :param old_dict: the dictionary to be compared with
        :param new_dict: the dictionary to compare 
        :param ignore_starting: keys to be ignored starting with string (default None)
        :param ignore_ending: keys to be ignored ending with string (default None)
        :param unnested_result: boolean if return unnest(normalized) result
        :param separator: if unnested_result separator used to normalize keys
        """
        if old_dict == new_dict:
            return
        result = {}
        try:
            old = self.exclude_keys(old_dict.keys(), ignore_starting, ignore_ending)
            new = self.exclude_keys(new_dict.keys(), ignore_starting, ignore_ending)
            overlapping_keys = old & new
            old_keys = old - new
            new_keys = new - old
            if old_keys:
                for key in old_keys:
                    result.update({key: {'action': 'del', 'old': old_dict[key]}})
            if new_keys:
                for key in new_keys:
                    result.update({key: {'action': 'add', 'new': new_dict[key]}})
            for key in overlapping_keys:
                if isinstance(old_dict[key], list) and isinstance(new_dict[key], list):
                    old_d = self.listtodict(old_dict[key])
                    new_d = self.listtodict(new_dict[key])
                    value = self.deepcompare(old_d, new_d)
                    if value is None:
                        continue
                else:
                    value = self.deepcompare(old_dict[key], new_dict[key])
                if value is None:
                    continue
                if all(f'{index}' == key for index, key in enumerate(sorted(value.keys()))):
                    value = [v for _, v in value.items()]
                result.update({key: value})
        except AttributeError:
            if old_dict != new_dict:
                result = {'action': 'mod', 'old': old_dict, 'new': new_dict}
        finally:
            if result:
                return result

    def flat(self, dictionary=None, separator='.', unpack_lists=False):
        if unpack_lists == True:
            return self.unnest(self.listtodict(dictionary or self.result), separator=separator)
        else:
            return self.unnest(dictionary or self.result, separator=separator)


import unittest
import ijr.firefly as ff
import os
import warnings



class TestFF(unittest.TestCase):

    def test_(self):
        warnings.filterwarnings(action="ignore", message="unclosed", category=ResourceWarning)
        x = ff.Secrets()
        y = x.dict_secret(os.environ['MONGO'])
        self.assertIsInstance(y, dict)


if __name__ == '__main__':
    unittest.main()



"""for secret manager """
from google.cloud import secretmanager_v1 as sm
from types import SimpleNamespace
from json import loads as jloads
from ijr.generic_lib import running_in_gcf
from os import environ
from ijr.mongo_lib import MongoReader


class NestedNamespace(SimpleNamespace):
    """need to access list indexs by index id
         {"test_array": [{"test": "index value 0"}, {"test": "index value 1"}]}
         test_array.i0.test  will return the value of the test node in index 0 of test_array
         so "index value 0" """

    def __init__(self, dictionary, **kwargs):
        super().__init__(**kwargs)
        for key, value in dictionary.items():
            if isinstance(value, dict):
                self.__setattr__(key, NestedNamespace(value))
            elif isinstance(value, list):
                list_dict = {f"i{str(index)}": _value
                             for index, _value in enumerate(value)}
                self.__setattr__(key, NestedNamespace(list_dict))
            else:
                self.__setattr__(key, value)


class Secrets:

    def __init__(self, project_id=None):
        scraped_id = environ.get('GCP_PROJECT', project_id)
        self.project_id = scraped_id
        if running_in_gcf():
            self.client = sm.SecretManagerServiceClient()
        else:
            import logging
            logging.warning('SecretManager -> Running local; using ./account.json')
            self.client =sm.SecretManagerServiceClient.from_service_account_json('account.json')

    def dict_secret(self, secret_id, version_id=None):
        """
        Access the payload for the given secret version if one exists. The version
        can be a version number as a string (e.g. "5") or an alias (e.g. "latest").
        """
        if version_id is None:
            version_id = "latest"
            # Build the resource name of the secret version.
        name = f"projects/{self.project_id}/secrets/{secret_id}/versions/{version_id}"
        secret = self.client.access_secret_version(name=name)
        _payload = secret.payload.data.decode('UTF-8')
        payload = jloads(_payload)
        return payload

    def dot_secret(self, secret_id, version_id=None):
        """
          Access the payload for the given secret version if one exists. The version
          can be a version number as a string (e.g. "5") or an alias (e.g. "latest").
          Returns a namespace for "dot" access
          """

        payload = self.dict_secret(secret_id, version_id=None)
        if payload:
            return NestedNamespace(payload)


class Config:
    """defaults to a function running with an od variable 'MONGO' """
    DEFAULT_DB_NAME = "common"
    DEFAULT_COLLECTION_NAME = "configs"
    DEFAULT_SECRET_TARGET = "MONGO"

    def __init__(self, project_id=None, secret_target=None, secret_id=None):
        if secret_id is None:
            secret_id = self._get_secret_id_from_env_var(secret_target)
        self._function_name = environ.get('FUNCTION_NAME')
        secrets = Secrets(project_id)
        self.mongo = secrets.dot_secret(secret_id)

    def _get_secret_id_from_env_var(self, secret_target=None):
        if secret_target is None:
            secret_target = self.DEFAULT_SECRET_TARGET
        secret_id = environ[str(secret_target)]
        return secret_id

    def get_configs_dict(self, _id=None, db_name=None, collection_name=None):
        if _id is None:
            _id = self._function_name
        if db_name is None:
            db_name = self.DEFAULT_DB_NAME
        if collection_name is None:
            collection_name = self.DEFAULT_COLLECTION_NAME
        with MongoReader(mdb_server=self.mongo.MDB_SERVER,
                         mdb_user=self.mongo.MDB_USER,
                         mdb_pass=self.mongo.MDB_PASS) as mr_configs:
            doc_list = mr_configs.find(db_name=db_name,
                                       collection_name=collection_name,
                                       query={"_id": _id})
        return next(doc_list, None)

    def get_configs_dot(self, _id=None, db_name=None, collection_name=None):
        temp = self.get_configs_dict(_id=_id,
                                     db_name=db_name,
                                     collection_name=collection_name)
        if temp:
            return NestedNamespace(temp)
        return None



from datetime import datetime

from pymongo import MongoClient, ReplaceOne, InsertOne, UpdateOne, UpdateMany


class MongoWriter(object):
    """A MongoClient for bulk writing operations"""

    def __init__(self, mdb_server, mdb_user, mdb_pass, db_name, col_name, threshold=250):
        self.db_name = db_name
        self.col_name = col_name
        self._statements = list()
        # Connect with 3.4 connection-string to Atlas -> user:pwd@server....
        self._client = MongoClient('mongodb://%s:%s@%s' % (mdb_user, mdb_pass, mdb_server))
        self._threshold = threshold
        self._write_counter = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _write_to_server(self):
        """bulk write statements to the server"""
        if not self._statements:
            return
        db = self._client.get_database(self.db_name)
        coll = db.get_collection(self.col_name)
        coll.bulk_write(self._statements)
        self._statements.clear()
        self._write_counter += 1

    def close(self):
        """write statements and disconnect from MongoDB."""
        self._write_to_server()
        self._client.close()

    def write_data(self, doc: dict, doc_key: str = None):
        """write document with _ts (timestamp) included

        :Parameters:
         - `doc`: A document to be written
         - `doc_key` (optional): Document key (_id) to be used for 
         document replacement/upsert
        """
        doc['_ts'] = datetime.now()
        if doc_key:
            doc['_id'] = '%s' % doc_key
            self._statements.append(ReplaceOne(filter={'_id': doc['_id']}, 
                                               replacement=doc, 
                                               upsert=True))
        else:
            self._statements.append(InsertOne(document=doc))
        if len(self._statements) > self._threshold:
            self._write_to_server()
        return self._write_counter

    def edit_data(self, query: dict, field: dict, mode: str):
        """edit document

        :Parameters:
         - `query`: A query that matches the document to update.
         - `field`: The modifications to apply.
         - `mode`: Editing mode 'one' or 'many'
        """
        if not isinstance(query, dict):
            raise TypeError('query must be an instance of dict')

        if not isinstance(field, dict):
            raise TypeError('field must be an instance of dict')

        if not isinstance(mode, str):
            raise TypeError('mode must be an instance of str')

        modes = {"one", "many"}
        mode = mode.lower()
        if mode not in modes:
            raise ValueError('Invalid mode. Expected one of: {"one", "many"}')

        if mode == "one":
            self._statements.append(UpdateOne(filter=query,
                                              update={'$set': field}))

        elif mode == "many":
            self._statements.append(UpdateMany(filter=query,
                                               update={'$set': field}))

        if len(self._statements) > self._threshold:
            self._write_to_server()


class MongoReader(object):
    def __init__(self, mdb_server, mdb_user, mdb_pass):
        self._client = MongoClient('mongodb://%s:%s@%s' % (mdb_user, mdb_pass, mdb_server))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._client.close()

    def find(self, db_name, collection_name, query, sorting=None, limit=-1, projection=None):
        db = self._client.get_database(db_name)
        col = db.get_collection(collection_name)
        ret = col.find(query, projection)
        if sorting:
            ret.sort(sorting)
        if limit > 0:
            ret.limit(limit=limit)
        for r in ret:
            yield r

    def aggregate(self, db_name, collection_name, pipeline):
        db = self._client.get_database(db_name)
        col = db.get_collection(collection_name)
        ret = col.aggregate(pipeline, allowDiskUse=True)
        for r in ret:
            yield r

    def collections(self, db_name):
        db = self._client.get_database(db_name)
        return db.list_collection_names()



from setuptools import setup



__version__ = '0.0.19'

setup(
    name='ijr',
    version=__version__,
    packages=['ijr'],
    url='https://github.com/IJsvogel/ijr',
    license='MIT',
    author='IT team IJsvogel Retail',
    author_email='it@ijsvogelretail.nl',
    description='IJsvogel Package',
    install_requires=['pymongo', 'google-cloud-pubsub', 'google-cloud-secret-manager']
)
